Continuing from [[3.07 Docker Compose with multiple services]], we will take this further by introducing a third service that communicates with these two services to combine their timestamps into one page. That would prove internal communication between containers.

Use the same folder structure as the previous challenge. So like this:
![[Pasted image 20250304002058.png]]

Modify docker-compose.yml to:
```
version: "3.8"
services:
  express-app1:
    image: siphon880d/express-app:latest
    ports:
      - "3001:3000"
    volumes:
      - ./logs1:/app/logs
  express-app2:
    image: siphon880d/express-app:latest
    ports:
      - "3002:3000"
    volumes:
      - ./logs2:/app/logs
  express-combinator:
    build:
      context: .
    ports:
      - "3003:3000"
```

Explanation: We run multiple instances of express-app, having to name them as different containers aka services to Docker: express-app1, and express-app2. Each service is to port 3001 and port 3002 on the host computer.  We added a **third service** that will look for a Dockerfile at the current folder to build an image and run a virtual container off of that image (Recall that `context: .` means to look at the current folder for Dockerfile).

The idea of this third service is that by visiting at port 3003, you see the combined timestamps of port 3001 and port 3002's visits, proving that the third service internally communicated with sibling services.

This means we need to add additional files.

---

### Additional files

.dockerignore (Standard):
```
.DS_Store
node_modules
package-lock.json
__pycache__
*.pyc
.env
```

Dockerfile (Standard for running Express with or without other NodeJS packages):
```
# Use the specific NodeJS base image (bundled in Debian)
FROM node:18

# Set the working directory in the container.
# All below commands will work from working directory.
WORKDIR /app

# Copy the package.json file to the container
COPY package.json .

# Install the NodeJS packages
RUN npm install

# Copy your application files (if any)
COPY . .

# Specify the command to run your application (optional, depends on your use case)
CMD ["node", "app.js"]
```

package.json (Standard for running Express on NodeJS 18 - base image):
```
{
    "dependencies": {
        "express": "4.21.2"
    }
}
```

app.js:
```
const express = require('express');
const app = express();

app.get('/', (req, res) => {
  res.json({ message: "Not a timestamp recorder. This is a combinator. Visit /visited to combine :3001 and :3002's timestamps!" });
});

app.get('/visited', (req, res) => {
  Promise.all([
    fetch('http://express-app1:3000/visited'),
    fetch('http://express-app2:3000/visited')
  ])
  .then(responses => Promise.all(responses.map(r => r.text())))
  .then(timestamps => {
    const combined = timestamps.join('<br/>');
    res.send(combined);
  })
  .catch(error => {
    console.error('Error fetching timestamps:', error);
    res.status(500).send('Error combining timestamps');
  });
});

const port = 3000;
app.listen(port, () => {
  console.log(`Server running on port ${port}`);
});
```

**Key takeaways from app.js:**
Notice that when the user visits /visited (on the computer web browser, the user visits localhost:3003/visited), they see the concatenation of timestamps from port 3001 visits and port 3002 visits. App.js makes fetch requests to port 3000 for both express-app1 and express-app2 because they're communicating internally across the Docker container network. 

In your `docker-compose.yml`, although you have two services: `express-app1` and `express-app2`, both running the same image but on different host ports (`3001` and `3002`) -- Since they are part of the same Docker network created automatically by Docker Compose (`docker compose up -d`), they can communicate with each other using their **service names** as hostnames. At the container level, since they are both port 3000, the fetches from express-app-combinator are:
```
  Promise.all([
    fetch('http://express-app1:3000/visited'),
    fetch('http://express-app2:3000/visited')
  ])
```

Your VS Code File Navigator looks like:
![[Pasted image 20250304010649.png]]

---

Run the docker-compose.yml:
```
docker compose up -d
```

Visit `:3001/` in the web browser to record a timestamp at service #1.
Visit `:3002/` in the web browser to record a timestamp at service #2.

Then visit `:3001/visited` and `:3002/visited` to look at their timestamps.

Visit `:3003/visited` in the web browser to see the combinations of both services' timestamps.

---

It may look like this:

3001/visited:
```
2025-03-04T08:48:43.687Z
```

3002/visited:
```
2025-03-04T08:48:46.851Z
```


3003/visited:
```
2025-03-04T08:48:43.687Z

2025-03-04T08:48:46.851Z
```

Or:
![[Pasted image 20250304010553.png]]

**Cleanup**
When done, perform cleanup. They've been in detached/daemon/background mode.

---

## Introspection: Backend vs Frontend Communications


Let's refer back to the above docker-compose.yml:
```
version: "3.8"
services:
  express-app1:
    image: siphon880d/express-app:latest
    ports:
      - "3001:3000"
    volumes:
      - ./logs1:/app/logs
  express-app2:
    image: siphon880d/express-app:latest
    ports:
      - "3002:3000"
    volumes:
      - ./logs2:/app/logs
  express-combinator:
    build:
      context: .
    ports:
      - "3003:3000"
```

If express-app1 makes a fetch request in the backend (inside the express js script):
```
http://express-app2:3000
```

And from inside `express-app2`, you can reach `express-app1` in the backend using:
```
http://express-app1:3000
```

However, if you visited an endpoint on the web browser that showed a webpage that can make request to the other service, let’s say you visited in the web browser `localhost:3001/gui`  which let’s say is HTML and Javascript responded by express-app1, and it shows a button to get timestamps from express-app2, on the frontend, that frontend javascript must make a call to express-app2 using the hosted port, NOT the containerized port:
```
http://localhost:3002
```

This is because the origin of the request is on the HOST computer through the host’s web browser.